-- localize some functions for performance
local ceil = math.ceil
local targetIsRoid = PR.TargetIsRoid
local isScanned = PR.IsScanned
local getRadarSelectionID = radar.GetRadarSelectionID
local setRadarSelection = radar.SetRadarSelection
local processGKCommand = gkinterface.GKProcessCommand

-- disable auto-scanning and single target search
function PR.ScannerOffSwitch()
    if PR.AutoScanActive then
        PR.Print("Auto-scanning deactivated.")
        PR.AutoScanActive = false
        PR.Interop.SetTargetlessLock(false)
    end
    PR.SingleSearchStop()
    PR.GUI.ScanTop30Idx = false
    PR.MassTestActive = false
    if PR.AutoScanTimer:IsActive() then PR.AutoScanTimer:Kill() end
    if PR.SingleSearchTimer:IsActive() then PR.SingleSearchTimer:Kill() end
    setRadarSelection()
end

-- tries to target PR.SingleSearchTarget at given node,
-- returns distance to target if successful
local function targetRoidCheck(node, id)
    setRadarSelection(node, id)
    local targetNode, targetId = getRadarSelectionID()
    local dist = targetIsRoid()
    if targetNode == node and targetId == id and dist ~= nil then
        return dist
    end
end

function PR.TryToTarget(targetId)
    local dist
    if PR.RoidNode > 0 then
        dist = targetRoidCheck(PR.RoidNode, targetId)
    else
        -- asteroids seem to be either nodeid 1 or 2, so limit search to those
        -- two for efficiency reasons.
        dist = targetRoidCheck(1, targetId)
        if not dist then
            dist = targetRoidCheck(2, targetId)
        end
    end
    return dist
end

-- repeatedly try to target PR.SingleSearchTarget
function PR.SingleSearch()
    if PR.SingleSearchTarget == 0 then return end
    if PR.SingleSearchTimer:IsActive() then PR.SingleSearchTimer:Kill() end

    local dist = PR.TryToTarget(PR.SingleSearchTarget)
    if not dist then
        setRadarSelection()
        if PR.Config.VerbosityLevel > 1 then
            PR.Print("Asteroid #" .. PR.SingleSearchTarget ..
                " is not within radar range")
        end
        PR.SingleSearchTimer:SetTimeout(5000, PR.SingleSearch)
    else
        -- target found
        PR.SingleSearchStop()
    end
end

-- gets called when clicking on an asteroid in the GUI's sector view
function PR.SingleSearchStart(roidId)
    if PR.SingleSearchTarget == roidId and PR.SingleSearchTimer:IsActive() then
        return
    end
    setRadarSelection()
    PR.SingleSearchTarget = roidId
    PR.SingleSearch()
end

-- ends single asteroid search and adjusts GUI
-- if @notify is passed and single search is active then print a status message
function PR.SingleSearchStop(notify)
    if PR.SingleSearchTarget > 0 then
        if PR.SingleSearchTimer:IsActive() then PR.SingleSearchTimer:Kill() end
        if notify then
            PR.Print("Search for asteroid #" .. PR.SingleSearchTarget ..
                " cancelled.")
        end
        PR.SingleSearchTarget = 0
        if PR.GUI.View == "Sector" then
            PR.GUI.Btn_Top4.title = PR.MiningBeamEquipped and
                "Start Top30 Check" or ""
            PR.GUI.Btn_Top4.fgcolor = PR.GUI.Color.Button
        end
    end
end

local function getNextAsteroid()
    local closestRoid = { 0, 0, 99999 } -- { node, id, dist }
    local loopcount = 0 -- prevent infinite loop, just in case
    -- Get a first radar target as a starting point for cycling through
    processGKCommand("RadarNext")
    local targetNode, targetId = getRadarSelectionID()
    local firstTarget = targetId
    -- Cycle through targets until we reach the first target again.
    -- Every cycle contains a nil target, that's why we need to RadarNext
    -- at least twice before declaring a system empty.
    -- Note that theoretically the first target could be out of range by
    -- the time we get back to it but in testing I was unable to create
    -- such a situation, the whole process is just too fast
    repeat
        loopcount = loopcount + 1
        processGKCommand("RadarNext")
        targetNode, targetId = getRadarSelectionID()
        if targetNode and targetId then
            local dist = targetIsRoid()
            if dist then
                dist = ceil(dist)
                if not isScanned(targetId) then
                    if dist <= PR.ScanRange then
                        -- if target in scanner range, go for it right away
                        -- even if others are closer
                        return { targetNode, targetId }
                    elseif dist < closestRoid[3] then
                        -- if none are in scanner range then find the closest one
                        closestRoid = { targetNode, targetId, dist }
                    end
                end
            end
        end
    until (targetId == firstTarget or loopcount > 1000)
    return closestRoid
end

-- automatically target unscanned asteroids for scanning
function PR.AutoScan()
    if not PR.AutoScanActive or PR.AutoScanTimer:IsActive() then return end

    local targetRoid = getNextAsteroid()
    if targetRoid[2] ~= 0 then
        setRadarSelection(targetRoid[1], targetRoid[2])
    else
        -- no unknown asteroid in range; clear target and try again later
        setRadarSelection()
        PR.AutoScanTimer:SetTimeout(PR.Config.AutoScanFrequency,
            PR.AutoScan)
    end
end

-- toggle auto-scanning
function PR.AutoScanToggle()
    -- stop other scanners
    PR.SingleSearchStop(true)
    PR.GUI.ScanTop30Idx = false

    if PR.AutoScanActive then
        PR.ScannerOffSwitch()
    elseif not PR.PluginReady then
        PR.Print("Auto-scanning failed: Prospector is not ready yet")
    elseif PR.Data.Unscanned == 0 and PR.Data.MaxId > 0 then
        PR.Print("Auto-scanning failed: All known " .. PR.Data.MaxId ..
            " asteroids in the sector have already been scanned")
    elseif not PR.ScannerEquipped then
        PR.Print("Auto-scanning failed: No equipped mineral scanner detected")
    elseif PlayerInStation() then
        PR.Print("Auto-scanning failed: Undock your ship first")
    else
        PR.AutoScanActive = true
        PR.StatusPrint(true)
        PR.Interop.SetTargetlessLock(true)
        PR.AutoScan()
    end
end
